==> commonc++-1.9.0.tar.gz: CommonC++-1.9.0/src/url.cpp <==
// Copyright (C) 2001 Open Source Telecom Corporation.
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// 
// As a special exception to the GNU General Public License, permission is 
// granted for additional uses of the text contained in its release 
// of Common C++.
// 
// The exception is that, if you link the Common C++ library with other
// files to produce an executable, this does not by itself cause the
// resulting executable to be covered by the GNU General Public License.
// Your use of that executable is in no way restricted on account of
// linking the Common C++ library code into it.
//
// This exception does not however invalidate any other reasons why
// the executable file might be covered by the GNU General Public License.
// 
// This exception applies only to the code released under the 
// name Common C++.  If you copy code from other releases into a copy of
// Common C++, as the General Public License permits, the exception does
// not apply to the code that you add in this way.  To avoid misleading
// anyone as to the status of such modified files, you must delete
// this exception notice from them.
// 
// If you write modifications of your own for Common C++, it is your choice
// whether to permit this exception to apply to your modifications.
// If you do not wish that, delete this exception notice.  

#include "config.h"
#include "export.h"
#include "strchar.h"
#include "exception.h"
#include "thread.h"
#include "socket.h"
#include "url.h"

#include <string>
#include <cstdlib>
#include <fcntl.h>
#include <cerrno>
#include <iostream>
#ifdef	WIN32
#include <io.h>
#endif
#include <strstream>
#include <ctype.h>

#ifdef	__NAMESPACES__
namespace ost {
using namespace std;
#endif

URLStream::URLStream() :
TCPStream()
{
	persistent = false;
	proxyPort = 0;
	timeout = 0;
	protocol = HTTP_PROTOCOL_1_0;
	follow = true;
	encoding = URL_BINARY_ENCODING;
	auth = URL_ANONYMOUS_AUTH;
	cookie = agent = pragma = referer = user = password = NULL;
	setError(false);
}

int URLStream::underflow(void)
{
	int len, rlen;
	char *buf;

	if(bufsize == 1)
		return TCPStream::underflow();

	if(!gptr())
		return EOF;

	if(gptr() < egptr())
		return (unsigned char)*gptr();

	rlen = (gbuf + bufsize) - eback();
	if(encoding == URL_CHUNKED_ENCODING)
	{
		buf = (char *)eback();
		*buf = '\n';
		while(!chunk && (*buf == '\n' || *buf == '\r'))
		{
			*buf = 0;
			len = Readline(buf, rlen, timeout);
		}
		if(len)
		{
			if(!chunk)
				chunk = strtol(buf, NULL, 16);

			if(rlen > (int)chunk)
				rlen = chunk;
		}
		else
			rlen = -1;
	}

	if(rlen > 0)
	{
		if(Socket::state == SOCKET_STREAM)
			rlen = ::read(so, (char *)eback(), rlen);
		else if(timeout)
		{
			if(Socket::isPending(SOCKET_PENDING_INPUT, timeout))
				rlen = ::recv(so, (char *)eback(), rlen, 0);
			else
				rlen = -1;
		}
		else
			rlen = ::recv(so, (char *)eback(), rlen, 0);
	}
	if(encoding == URL_CHUNKED_ENCODING && rlen > 0)
		chunk -= rlen;

	if(rlen < 1)
	{
		if(rlen < 0)
			clear(ios::failbit | rdstate());
		return EOF;
	}
	setg(eback(), eback(), eback() + rlen);
	return (unsigned char)*gptr();
}
	
void URLStream::setProxy(const char *host, tpport_t port)
{
	proxyHost = host;
	proxyPort = port;
}

urlerror_t URLStream::Submit(const char *path, const char **vars, int bufsize)
{
	char buffer[512];
	urlerror_t status = URL_INVALID;

	strncpy(buffer, path, sizeof(buffer) - 1);
	buffer[sizeof(buffer) - 1] = 0;
	path = buffer;

	if(!strnicmp(path, "http:", 5))
	{
		urlmethod = HTTP_GET_METHOD;
		path = strchr(path + 5, '/');
		status = sendHTTPHeader(path, vars, bufsize);
	}
	if(status)
	{
		if(Socket::state != SOCKET_AVAILABLE)
			Close();
	}
	else
		status = getHTTPHeaders(buffer, sizeof(buffer));
	return status;
}

urlerror_t URLStream::Post(const char *path, const char **vars, int bufsize)
{
	char buffer[512];
	urlerror_t status = URL_INVALID;

	strncpy(buffer, path, sizeof(buffer) - 1);
	buffer[sizeof(buffer) - 1] = 0;
	path = buffer;

	if(!strnicmp(path, "http:", 5))
	{
		urlmethod = HTTP_POST_METHOD;
		path = strchr(path + 5, '/');
		status = sendHTTPHeader(path, vars, bufsize);
	}
	if(status)
	{
		if(Socket::state != SOCKET_AVAILABLE)
			Close();
	}
	else
		status = getHTTPHeaders(buffer, sizeof(buffer));
	return status;
}


urlerror_t URLStream::Head(const char *path, int bufsize)
{
	char buffer[512];
	urlerror_t status = URL_INVALID;

	strncpy(buffer, path, sizeof(buffer) - 1);
	buffer[sizeof(buffer) - 1] = 0;
	path = buffer;

	if(!strnicmp(path, "http:", 5))
	{
		urlmethod = HTTP_GET_METHOD;
		path = strchr(path + 5, '/');
		status = sendHTTPHeader(path, NULL, bufsize);
	}
	if(status)
	{
		if(Socket::state != SOCKET_AVAILABLE)
			Close();
	}
	else
		status = getHTTPHeaders(buffer, sizeof(buffer));
	return status;
}

URLStream &URLStream::getline(char *buffer, size_t size)
{
	size_t len;

	*buffer = 0;
	iostream::getline(buffer, size);
	len = strlen(buffer);
	
	while(len)
	{
		if(buffer[len - 1] == '\r' || buffer[len - 1] == '\n')
			buffer[len - 1] = 0;
		else
			break;
		--len;
	}
	return *this;
}

urlerror_t URLStream::Get(const char *urlpath, int bufsize)
{
	const char *path = urlpath;
	char buffer[512];
	urlerror_t status;

	strncpy(buffer, path, sizeof(buffer) - 1);
	buffer[sizeof(buffer) - 1] = 0;
	path = buffer;
	urlmethod = FILE_GET_METHOD;

	if(Socket::state != SOCKET_AVAILABLE)
		Close();

	if(!strnicmp(path, "file:", 5))
	{
		urlmethod = FILE_GET_METHOD;
		path += 5;
	}
	else if(!strnicmp(path, "http:", 5))
	{
		urlmethod = HTTP_GET_METHOD;
		path = strchr(path + 5, '/');
	}
	switch(urlmethod)
	{
	case HTTP_GET_METHOD:
		status = sendHTTPHeader(path, NULL, bufsize);
		break;
	case FILE_GET_METHOD:
		if(so != INVALID_SOCKET)
			close(so);
		so = ::open(path, O_RDONLY);
                // FIXME: open return the same handle type as socket call ??
		if(so == INVALID_SOCKET)
			return URL_INVALID;
		Socket::state = SOCKET_STREAM;
		Allocate(bufsize);
		return URL_SUCCESS;
	}
	if(status)
	{
		if(Socket::state != SOCKET_AVAILABLE)
			Close();
	}
	else
		status = getHTTPHeaders(buffer, sizeof(buffer));
	return status;
}

urlerror_t URLStream::getHTTPHeaders(char *buffer, unsigned bufsize)
{
	urlerror_t status = URL_SUCCESS;
	char *cp, *ep, len = 1;

	chunk = ((unsigned)-1) / 2;
	encoding = URL_BINARY_ENCODING;
	while(len > 0)
	{
		len = Readline(buffer, bufsize, timeout);
		if(len < 1)
			status = URL_TIMEOUT;

		if(buffer[0] == ' ' || buffer[0] == '\r' || buffer[0] == '\n')
			break;
		cp = strchr(buffer, ':');
		if(!cp)
			continue;
		*(cp++) = 0;
		while(*cp == ' ' || *cp == '\t')
			++cp;
		ep = strchr(cp, '\n');
		if(!ep)
			ep = "";
		while(*ep == '\n' || *ep == '\r' || *ep == ' ')
		{
			*ep = 0;
			if((--ep) < cp)
				break;
		}
		if(!stricmp(buffer, "Transfer-Encoding"))
		{
			if(!stricmp(cp, "chunked"))
			{
				chunk = 0;
				encoding = URL_CHUNKED_ENCODING;
			}
		}	
		httpHeader(buffer, cp);
	}
	return status;
}


void URLStream::Close(void)
{
	if(Socket::state == SOCKET_AVAILABLE)
		return;

	endStream();
	so = socket(AF_INET, SOCK_STREAM, 0);
	if(so != INVALID_SOCKET)
		Socket::state = SOCKET_AVAILABLE;
}

urlerror_t URLStream::sendHTTPHeader(const char *url, const char **vars, int bufsize)
{
	char buffer[2048];
	char reloc[1024];
	char urlbuf[1024];
	char host[128];
	strstream str(buffer, sizeof(buffer), ios::out);
	char *ref, *cp, *ep;
	char *hp;
	InetHostAddress server;
	int count = 0;
	unsigned len = 0;
	tpport_t port;
	const char **args = vars;
	const char *var;

retry:
	buffer[0] = 0;
	str.seekp(0);
	strncpy(host, url, 127);
	host[127] = 0;
	hp = strchr(host, '/');
	while(*hp == '/')
		++hp;
	ep = strrchr(hp, ':');
	cp = strchr(hp, '/');
	if(ep)
		*ep = 0;
	else if(cp)
		*cp = 0;

	if(!proxyPort)
	{
		strncpy(urlbuf, url, 1024);
		ep = strchr(urlbuf, ':');
		if(ep)
		{
			*(ep++) = 0;
			port = atoi(ep);
		}
		else
			port = 80;
		while(*url == '/')
			++url;
		const char* ep = strchr(url, '/');
		if(ep)
	        	strncpy(urlbuf, ep, sizeof(urlbuf) - 1);
		else
			strcpy(urlbuf, "/");
                urlbuf[sizeof(urlbuf) - 1] = 0;
                url = urlbuf;
	}

	switch(urlmethod)
	{
	case HTTP_GET_METHOD:
		str << "GET ";
		if(proxyPort)
			str << "http:" << url;
		else
			str << url;
		if(!cp && proxyPort)
			str << '/';
		break;
	case HTTP_POST_METHOD:
		str << "POST ";
		if(proxyPort)
			str << "http:" << url;
		else
			str << url;
		if(!cp && proxyPort)
			str << '/';
		break;
	default:
		return URL_INVALID;
	}
	if(port && port != 80 && proxyPort)
		str << ":" << port;

	if(vars && urlmethod == HTTP_GET_METHOD)
	{
		str << "?";
		while(*vars)
		{
			if(count++)
				str << "&";
			str << *vars;
			++vars;
		}
	}

	switch(protocol)
	{
	case HTTP_PROTOCOL_1_1:
		str << " HTTP/1.1" << "\r\n";
		break;
	case HTTP_PROTOCOL_1_0:
		str << " HTTP/1.0" << "\r\n";
		break;
	}

	str << "Host: " << hp << "\r\n";
        if(agent)
                str << "User-Agent: " << agent << "\r\n";

        if(cookie)
                str << "Cookie: " << cookie << "\r\n";

        if(pragma)
                str << "Pragma: " << pragma << "\r\n";

        if(referer)
                str << "Referer: " << referer << "\r\n";

        str << "Connection: close\r\n";
	if(vars && urlmethod == HTTP_POST_METHOD)
	{
		while(*args)
		{
			var = *args;
			if(count++)
				len += strlen(var) + 1;
			else
				len = strlen(var);
			++args;
		}
		count = 0;
		len += 2;
		str << "Content-Length: " << len << "\r\n";
	}
        str << "\r\n" << ends;

        if(Socket::state != SOCKET_AVAILABLE)
                        Close();

	if(proxyPort)
	        Connect(proxyHost, proxyPort, bufsize);
	else
	{
		server = hp;
		Connect(server, port, bufsize);
	}

        if(!isConnected())
                return URL_UNREACHABLE;

        ::write(so, buffer, strlen(buffer));
	if(urlmethod == HTTP_POST_METHOD && vars)
	{
		str.seekp(0);
		count = 0;
		while(*vars)
		{
			if(count++)
				::write(so, "&", 1);
			var = *vars;
			::write(so, var, strlen(var));
			++vars;
		}
		::write(so, "\r\n", 2);
	}

cont:
        len = Readline(buffer, sizeof(buffer) - 1, timeout);
	if(len < 1)
		return URL_TIMEOUT;

        if(strnicmp(buffer, "HTTP/", 5))
                return URL_INVALID;

        ref = strchr(buffer, ' ');
        while(*ref == ' ')
                ++ref;

        switch(atoi(ref))
        {
        default:
                return URL_INVALID;
	case 100:
		goto cont;
        case 200:
                return URL_SUCCESS;
        case 401:
                return URL_UNAUTHORIZED;
        case 403:
                return URL_FORBIDDEN;
        case 404:
                return URL_MISSING;
        case 405:
                return URL_DENIED;
        case 500:
        case 501:
        case 502:
        case 503:
        case 504:
        case 505:
                return URL_FAILURE;
        case 300:
        case 301:
        case 302:
                break;
        }
	if(!follow)
		return URL_RELOCATED;
        for(;;)
        {
                len = Readline(reloc, sizeof(reloc), timeout);
		if(len < 1)
			return URL_TIMEOUT;
                if(!strnicmp(reloc, "Location: ", 10))
                        break;
        }
        if(!strnicmp(reloc + 10, "http:", 5))
        {
                url = strchr(reloc + 15, '/');
                ep = (char *)(url + strlen(url) - 1);
                while(*ep == '\r' || *ep == '\n')
                        *(ep--) = 0;
        }
        else
                url = reloc + 10;
        Close();
        goto retry;
}

void URLStream::setAuthentication(urlauth_t a, const char *value)
{
}

void URLStream::setReferer(const char *str)
{
	char *ref;
	char host[128];

	referer = str;
	if(!str)
		return;
	
	ref = strchr((char *)str, '/');
	if(ref)
	{
		if(ref[1] == '/')
			ref += 2;
		else
			++ref;
	}
	else
		ref = (char *)str;

	memcpy(host, ref, sizeof(host));
	host[sizeof(host) - 1] = 0;

	ref = strchr(ref, '/');
	if(ref)
		*ref = 0;

	refhost = host;
};

CCXX_EXPORT(char*) urldecode(char *source, char *dest)
{
	char *ret;

	if(!dest)
		dest = source;

	ret = dest;

	while(*source)
	{
		switch(*source)
		{
		case '+':
			*(dest++) = ' ';
			break;
		case '%':
			dest[0] = source[1];
			dest[1] = source[2];
			dest[2] = 0;
			*(dest++) = (char)strtol(dest, NULL, 16);
			source += 2;
			break;
		default:
			*(dest++) = *source;
		}
		++source;
	}
	*dest = 0;
	return ret;
}	

CCXX_EXPORT(char*) urlencode(const char *source, char *dest, unsigned max)	
{
	static char *hex = "0123456789abcdef";
	unsigned len = 0;
	unsigned char ch;
	char *ret = dest;

	while(len < max - 4 && *source)
	{
		ch = (unsigned char)*source;
		if(*source == ' ')
			*(dest++) = '+';
		else if(isalnum(*source) || *source == '/' || *source == '.')
			*(dest++) = *source;
		else
		{
			*(dest++) = '%';
			*(dest++) = hex[ch / 16];
			*(dest++) = hex[ch % 16];
		}	
		++source;
	}
	*dest = 0;
	return ret;
}

#ifdef	__NAMESPACES__
};
#endif

==> commonc++-1.9.2.tar.gz: CommonC++-1.9.2/src/url.cpp <==
// Copyright (C) 2001 Open Source Telecom Corporation.
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// 
// As a special exception to the GNU General Public License, permission is 
// granted for additional uses of the text contained in its release 
// of Common C++.
// 
// The exception is that, if you link the Common C++ library with other
// files to produce an executable, this does not by itself cause the
// resulting executable to be covered by the GNU General Public License.
// Your use of that executable is in no way restricted on account of
// linking the Common C++ library code into it.
//
// This exception does not however invalidate any other reasons why
// the executable file might be covered by the GNU General Public License.
// 
// This exception applies only to the code released under the 
// name Common C++.  If you copy code from other releases into a copy of
// Common C++, as the General Public License permits, the exception does
// not apply to the code that you add in this way.  To avoid misleading
// anyone as to the status of such modified files, you must delete
// this exception notice from them.
// 
// If you write modifications of your own for Common C++, it is your choice
// whether to permit this exception to apply to your modifications.
// If you do not wish that, delete this exception notice.  

#include "config.h"
#include "export.h"
#include "strchar.h"
#include "exception.h"
#include "thread.h"
#include "socket.h"
#include "url.h"

#include <string>
#include <cstdlib>
#include <fcntl.h>
#include <cerrno>
#include <iostream>
#ifdef	WIN32
#include <io.h>
#endif
#include <strstream>
#include <ctype.h>

#ifdef	CCXX_NAMESPACES
namespace ost {
using namespace std;
#endif

URLStream::URLStream() :
TCPStream()
{
	persistent = false;
	proxyPort = 0;
	timeout = 0;
	protocol = HTTP_PROTOCOL_1_0;
	follow = true;
	encoding = URL_BINARY_ENCODING;
	auth = URL_ANONYMOUS_AUTH;
	cookie = agent = pragma = referer = user = password = NULL;
	setError(false);
}

int URLStream::underflow(void)
{
	int len, rlen;
	char *buf;

	if(bufsize == 1)
		return TCPStream::underflow();

	if(!gptr())
		return EOF;

	if(gptr() < egptr())
		return (unsigned char)*gptr();

	rlen = (gbuf + bufsize) - eback();
	if(encoding == URL_CHUNKED_ENCODING)
	{
		buf = (char *)eback();
		*buf = '\n';
		while(!chunk && (*buf == '\n' || *buf == '\r'))
		{
			*buf = 0;
			len = Readline(buf, rlen, timeout);
		}
		if(len)
		{
			if(!chunk)
				chunk = strtol(buf, NULL, 16);

			if(rlen > (int)chunk)
				rlen = chunk;
		}
		else
			rlen = -1;
	}

	if(rlen > 0)
	{
		if(Socket::state == SOCKET_STREAM)
			rlen = ::read(so, (char *)eback(), rlen);
		else if(timeout)
		{
			if(Socket::isPending(SOCKET_PENDING_INPUT, timeout))
				rlen = ::recv(so, (char *)eback(), rlen, 0);
			else
				rlen = -1;
		}
		else
			rlen = ::recv(so, (char *)eback(), rlen, 0);
	}
	if(encoding == URL_CHUNKED_ENCODING && rlen > 0)
		chunk -= rlen;

	if(rlen < 1)
	{
		if(rlen < 0)
			clear(ios::failbit | rdstate());
		return EOF;
	}
	setg(eback(), eback(), eback() + rlen);
	return (unsigned char)*gptr();
}
	
void URLStream::setProxy(const char *host, tpport_t port)
{
	proxyHost = host;
	proxyPort = port;
}

urlerror_t URLStream::Submit(const char *path, const char **vars, int bufsize)
{
	char buffer[512];
	urlerror_t status = URL_INVALID;

	strncpy(buffer, path, sizeof(buffer) - 1);
	buffer[sizeof(buffer) - 1] = 0;
	path = buffer;

	if(!strnicmp(path, "http:", 5))
	{
		urlmethod = HTTP_GET_METHOD;
		path = strchr(path + 5, '/');
		status = sendHTTPHeader(path, vars, bufsize);
	}
	if(status)
	{
		if(Socket::state != SOCKET_AVAILABLE)
			Close();
	}
	else
		status = getHTTPHeaders(buffer, sizeof(buffer));
	return status;
}

urlerror_t URLStream::Post(const char *path, const char **vars, int bufsize)
{
	char buffer[512];
	urlerror_t status = URL_INVALID;

	strncpy(buffer, path, sizeof(buffer) - 1);
	buffer[sizeof(buffer) - 1] = 0;
	path = buffer;

	if(!strnicmp(path, "http:", 5))
	{
		urlmethod = HTTP_POST_METHOD;
		path = strchr(path + 5, '/');
		status = sendHTTPHeader(path, vars, bufsize);
	}
	if(status)
	{
		if(Socket::state != SOCKET_AVAILABLE)
			Close();
	}
	else
		status = getHTTPHeaders(buffer, sizeof(buffer));
	return status;
}


urlerror_t URLStream::Head(const char *path, int bufsize)
{
	char buffer[512];
	urlerror_t status = URL_INVALID;

	strncpy(buffer, path, sizeof(buffer) - 1);
	buffer[sizeof(buffer) - 1] = 0;
	path = buffer;

	if(!strnicmp(path, "http:", 5))
	{
		urlmethod = HTTP_GET_METHOD;
		path = strchr(path + 5, '/');
		status = sendHTTPHeader(path, NULL, bufsize);
	}
	if(status)
	{
		if(Socket::state != SOCKET_AVAILABLE)
			Close();
	}
	else
		status = getHTTPHeaders(buffer, sizeof(buffer));
	return status;
}

URLStream &URLStream::getline(char *buffer, size_t size)
{
	size_t len;

	*buffer = 0;
	iostream::getline(buffer, size);
	len = strlen(buffer);
	
	while(len)
	{
		if(buffer[len - 1] == '\r' || buffer[len - 1] == '\n')
			buffer[len - 1] = 0;
		else
			break;
		--len;
	}
	return *this;
}

urlerror_t URLStream::Get(const char *urlpath, int bufsize)
{
	const char *path = urlpath;
	char buffer[512];
	urlerror_t status;

	strncpy(buffer, path, sizeof(buffer) - 1);
	buffer[sizeof(buffer) - 1] = 0;
	path = buffer;
	urlmethod = FILE_GET_METHOD;

	if(Socket::state != SOCKET_AVAILABLE)
		Close();

	if(!strnicmp(path, "file:", 5))
	{
		urlmethod = FILE_GET_METHOD;
		path += 5;
	}
	else if(!strnicmp(path, "http:", 5))
	{
		urlmethod = HTTP_GET_METHOD;
		path = strchr(path + 5, '/');
	}
	switch(urlmethod)
	{
	case HTTP_GET_METHOD:
		status = sendHTTPHeader(path, NULL, bufsize);
		break;
	case FILE_GET_METHOD:
		if(so != INVALID_SOCKET)
			close(so);
		so = ::open(path, O_RDONLY);
                // FIXME: open return the same handle type as socket call ??
		if(so == INVALID_SOCKET)
			return URL_INVALID;
		Socket::state = SOCKET_STREAM;
		Allocate(bufsize);
		return URL_SUCCESS;
	}
	if(status)
	{
		if(Socket::state != SOCKET_AVAILABLE)
			Close();
	}
	else
		status = getHTTPHeaders(buffer, sizeof(buffer));
	return status;
}

urlerror_t URLStream::getHTTPHeaders(char *buffer, unsigned bufsize)
{
	urlerror_t status = URL_SUCCESS;
	char *cp, *ep, len = 1;
	char nc = 0;

	chunk = ((unsigned)-1) / 2;
	encoding = URL_BINARY_ENCODING;
	while(len > 0)
	{
		len = Readline(buffer, bufsize, timeout);
		if(len < 1)
			status = URL_TIMEOUT;

		if(buffer[0] == ' ' || buffer[0] == '\r' || buffer[0] == '\n')
			break;
		cp = strchr(buffer, ':');
		if(!cp)
			continue;
		*(cp++) = 0;
		while(*cp == ' ' || *cp == '\t')
			++cp;
		ep = strchr(cp, '\n');
		if(!ep)
			ep = &nc;
		while(*ep == '\n' || *ep == '\r' || *ep == ' ')
		{
			*ep = 0;
			if((--ep) < cp)
				break;
		}
		if(!stricmp(buffer, "Transfer-Encoding"))
		{
			if(!stricmp(cp, "chunked"))
			{
				chunk = 0;
				encoding = URL_CHUNKED_ENCODING;
			}
		}	
		httpHeader(buffer, cp);
	}
	return status;
}


void URLStream::Close(void)
{
	if(Socket::state == SOCKET_AVAILABLE)
		return;

	endStream();
	so = socket(AF_INET, SOCK_STREAM, 0);
	if(so != INVALID_SOCKET)
		Socket::state = SOCKET_AVAILABLE;
}

urlerror_t URLStream::sendHTTPHeader(const char *url, const char **vars, int bufsize)
{
	char buffer[2048];
	char reloc[1024];
	char urlbuf[1024];
	char host[128];
	strstream str(buffer, sizeof(buffer), ios::out);
	char *ref, *cp, *ep;
	char *hp;
	InetHostAddress server;
	int count = 0;
	unsigned len = 0;
	tpport_t port;
	const char **args = vars;
	const char *var;

retry:
	buffer[0] = 0;
	str.seekp(0);
	strncpy(host, url, 127);
	host[127] = 0;
	hp = strchr(host, '/');
	while(*hp == '/')
		++hp;
	ep = strrchr(hp, ':');
	cp = strchr(hp, '/');
	if(ep)
		*ep = 0;
	else if(cp)
		*cp = 0;

	if(!proxyPort)
	{
		strncpy(urlbuf, url, 1024);
		ep = strchr(urlbuf, ':');
		if(ep)
		{
			*(ep++) = 0;
			port = atoi(ep);
		}
		else
			port = 80;
		while(*url == '/')
			++url;
		const char* ep = strchr(url, '/');
		if(ep)
	        	strncpy(urlbuf, ep, sizeof(urlbuf) - 1);
		else
			strcpy(urlbuf, "/");
                urlbuf[sizeof(urlbuf) - 1] = 0;
                url = urlbuf;
	}

	switch(urlmethod)
	{
	case HTTP_GET_METHOD:
		str << "GET ";
		if(proxyPort)
			str << "http:" << url;
		else
			str << url;
		if(!cp && proxyPort)
			str << '/';
		break;
	case HTTP_POST_METHOD:
		str << "POST ";
		if(proxyPort)
			str << "http:" << url;
		else
			str << url;
		if(!cp && proxyPort)
			str << '/';
		break;
	default:
		return URL_INVALID;
	}
	if(port && port != 80 && proxyPort)
		str << ":" << port;

	if(vars && urlmethod == HTTP_GET_METHOD)
	{
		str << "?";
		while(*vars)
		{
			if(count++)
				str << "&";
			str << *vars;
			++vars;
		}
	}

	switch(protocol)
	{
	case HTTP_PROTOCOL_1_1:
		str << " HTTP/1.1" << "\r\n";
		break;
	case HTTP_PROTOCOL_1_0:
		str << " HTTP/1.0" << "\r\n";
		break;
	}

	str << "Host: " << hp << "\r\n";
        if(agent)
                str << "User-Agent: " << agent << "\r\n";

        if(cookie)
                str << "Cookie: " << cookie << "\r\n";

        if(pragma)
                str << "Pragma: " << pragma << "\r\n";

        if(referer)
                str << "Referer: " << referer << "\r\n";

        str << "Connection: close\r\n";
	if(vars && urlmethod == HTTP_POST_METHOD)
	{
		while(*args)
		{
			var = *args;
			if(count++)
				len += strlen(var) + 1;
			else
				len = strlen(var);
			++args;
		}
		count = 0;
		len += 2;
		str << "Content-Length: " << len << "\r\n";
	}
        str << "\r\n" << ends;

        if(Socket::state != SOCKET_AVAILABLE)
                        Close();

	if(proxyPort)
	        Connect(proxyHost, proxyPort, bufsize);
	else
	{
		server = hp;
		Connect(server, port, bufsize);
	}

        if(!isConnected())
                return URL_UNREACHABLE;

        ::write(so, buffer, strlen(buffer));
	if(urlmethod == HTTP_POST_METHOD && vars)
	{
		str.seekp(0);
		count = 0;
		while(*vars)
		{
			if(count++)
				::write(so, "&", 1);
			var = *vars;
			::write(so, var, strlen(var));
			++vars;
		}
		::write(so, "\r\n", 2);
	}

cont:
        len = Readline(buffer, sizeof(buffer) - 1, timeout);
	if(len < 1)
		return URL_TIMEOUT;

        if(strnicmp(buffer, "HTTP/", 5))
                return URL_INVALID;

        ref = strchr(buffer, ' ');
        while(*ref == ' ')
                ++ref;

        switch(atoi(ref))
        {
        default:
                return URL_INVALID;
	case 100:
		goto cont;
        case 200:
                return URL_SUCCESS;
        case 401:
                return URL_UNAUTHORIZED;
        case 403:
                return URL_FORBIDDEN;
        case 404:
                return URL_MISSING;
        case 405:
                return URL_DENIED;
        case 500:
        case 501:
        case 502:
        case 503:
        case 504:
        case 505:
                return URL_FAILURE;
        case 300:
        case 301:
        case 302:
                break;
        }
	if(!follow)
		return URL_RELOCATED;
        for(;;)
        {
                len = Readline(reloc, sizeof(reloc), timeout);
		if(len < 1)
			return URL_TIMEOUT;
                if(!strnicmp(reloc, "Location: ", 10))
                        break;
        }
        if(!strnicmp(reloc + 10, "http:", 5))
        {
                url = strchr(reloc + 15, '/');
                ep = (char *)(url + strlen(url) - 1);
                while(*ep == '\r' || *ep == '\n')
                        *(ep--) = 0;
        }
        else
                url = reloc + 10;
        Close();
        goto retry;
}

void URLStream::setAuthentication(urlauth_t a, const char *value)
{
}

void URLStream::setReferer(const char *str)
{
	char *ref;
	char host[128];

	referer = str;
	if(!str)
		return;
	
	ref = strchr((char *)str, '/');
	if(ref)
	{
		if(ref[1] == '/')
			ref += 2;
		else
			++ref;
	}
	else
		ref = (char *)str;

	memcpy(host, ref, sizeof(host));
	host[sizeof(host) - 1] = 0;

	ref = strchr(ref, '/');
	if(ref)
		*ref = 0;

	refhost = host;
};

CCXX_EXPORT(char*) urldecode(char *source, char *dest)
{
	char *ret;

	if(!dest)
		dest = source;

	ret = dest;

	while(*source)
	{
		switch(*source)
		{
		case '+':
			*(dest++) = ' ';
			break;
		case '%':
			dest[0] = source[1];
			dest[1] = source[2];
			dest[2] = 0;
			*(dest++) = (char)strtol(dest, NULL, 16);
			source += 2;
			break;
		default:
			*(dest++) = *source;
		}
		++source;
	}
	*dest = 0;
	return ret;
}	

CCXX_EXPORT(char*) urlencode(const char *source, char *dest, unsigned max)	
{
	static const char *hex = "0123456789abcdef";
	unsigned len = 0;
	unsigned char ch;
	char *ret = dest;

	while(len < max - 4 && *source)
	{
		ch = (unsigned char)*source;
		if(*source == ' ')
			*(dest++) = '+';
		else if(isalnum(*source) || *source == '/' || *source == '.')
			*(dest++) = *source;
		else
		{
			*(dest++) = '%';
			*(dest++) = hex[ch / 16];
			*(dest++) = hex[ch % 16];
		}	
		++source;
	}
	*dest = 0;
	return ret;
}

#ifdef	CCXX_NAMESPACES
};
#endif
